/* SpamKiller for Qmail                                                       *
 * Author: Staffan Hml (staham@algonet.se)                                 *
 * Date started: 08-Dec-1997                                                  *
 * Version: 2.02                                                              *
 *
 * Description:
 *  This program is meant to be used via the .qmail files that the qmail      *
 * mail system uses to control what to do with incoming emails. This          *
 * program will give you the opportunity to control which mails you want      *
 * to receive and which are to be thrown away. It does this by checking       *
 * if the mail is directly addressed or not, and then checking various        *
 * configuration files to see what to do. The two most important files are    *
 * .allow and .deny that should reside in your home directory. .allow will    *
 * let the addresses defined in it through even if they are not directly      *
 * addressed. .deny will deny the addresses defined in it even if they have   *
 * mailed you directly. Instructions for usage can be found in the README     *
 * file.                                                                      *
 *                                                                            *
 * This program is free software; you can redistribute it and/or modify       *
 * it under the terms of the GNU General Public License as published by       *
 * the Free Software Foundation; either version 2 of the License, or          *
 * (at your option) any later version.                                        *
 *                                                                            *
 * This program is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with this program; if not, write to the Free Software                *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA    *
 *                                                                            *
 * Should you need to contact me, the author, you can do so by                *
 * e-mail - mail your message to <staham@algonet.se>.                         *
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#define TRUE 1
#define FALSE 0
#define FAULT_NOERR	0		/* No error. */
#define FAULT_NOTE	1		/* Just a notice message. */
#define FAULT_ERR		2		/* Filter failed. */
#define FAULT_KILL	3		/* Mail killed */

void fault(char *err,int errtype);
char * GetHeader(int index,int length);
int CheckDenyFile(void);
int CheckSubjectFile(void);
int CheckAllowFile(void);
void StrToLower(char *string);
void FreeAll(void);
char * GetSubject(void);
void PrintBounce(void);


time_t curtime;								/* Current time */
static char tstamp[80];						/* Buffer for timestamp */
static char tempfilename[150];			/* Temporary filename storage */

int iTo,iFrom,iCc,iBcc,iSubject;			/* Index values */
int lTo,lFrom,lCc,lBcc,lSubject;			/* Length values */
char *senderA,*recipientA,*home,*ext;	/* Environment variables */
char *sender,*recipient;
char *from,*to,*cc,*bcc,*subject;		/* Header fields */
int debug=FALSE,bounce=TRUE,shortlogs=FALSE,nologs=FALSE;

int main(void)
{
	static char tmpbuffer[1000];
	FILE *stream;
	int size,pos=0,error=FALSE,dir=0;
	char *tmpptr;

	curtime = time(0);  /* Get current timestamp */

	sender=0;recipient=0;
	senderA=(char *)getenv("SENDER");
	recipientA=(char *)getenv("RECIPIENT");
	home=(char *)getenv("HOME");
	ext=(char *)getenv("EXT");
	if(!ext) ext="";

	if(senderA)	sender=strdup(senderA);
	if(recipientA) recipient=strdup(recipientA);
	if(sender) StrToLower(sender);
	if(recipient) StrToLower(recipient);

	sprintf(tempfilename,"%s/.spamkiller-v2-tempfile-for-%s-%d",home,ext,(int)getpid());

	if (!(stream=fopen(tempfilename,"w")))
	{
		fault("Couldn't open temporary file for writing!",FAULT_ERR);
		return(0); /* Let the mail through! */
	}

	if(fgets(tmpbuffer,1000,stdin))
	{
		fputs(tmpbuffer,stream);
		size=strlen(tmpbuffer);

		while (TRUE)
		{
			if(tmpbuffer[0]=='\n') break;

			if(!strncasecmp(tmpbuffer,"From:",5))
			{
				lFrom=-5;
				iFrom=pos+5;
				do
				{
					lFrom+=size;
					pos+=size;
					if(tmpbuffer[size-1]!='\n') {error=TRUE;goto endofdata;}
					if(!fgets(tmpbuffer,1000,stdin)) goto endofdata;
					fputs(tmpbuffer,stream);
					size=strlen(tmpbuffer);
				}
				while (tmpbuffer[0]==' ' || tmpbuffer[0]=='\t');
			}
			else if(!strncasecmp(tmpbuffer,"To:",3))
			{
				lTo=-3;
				iTo=pos+3;
				do
				{
					lTo+=size;
					pos+=size;
					if(tmpbuffer[size-1]!='\n') {error=TRUE;goto endofdata;}
					if(!fgets(tmpbuffer,1000,stdin)) goto endofdata;
					fputs(tmpbuffer,stream);
					size=strlen(tmpbuffer);
				}
				while (tmpbuffer[0]==' ' || tmpbuffer[0]=='\t');
			}
			else if(!strncasecmp(tmpbuffer,"Cc:",3))
			{
				lCc=-3;
				iCc=pos+3;
				do
				{
					lCc+=size;
					pos+=size;
					if(tmpbuffer[size-1]!='\n') {error=TRUE;goto endofdata;}
					if(!fgets(tmpbuffer,1000,stdin)) goto endofdata;
					fputs(tmpbuffer,stream);
					size=strlen(tmpbuffer);
				}
				while (tmpbuffer[0]==' ' || tmpbuffer[0]=='\t');
			}
			else if(!strncasecmp(tmpbuffer,"Bcc:",4))
			{
				lBcc=-4;
				iBcc=pos+4;
				do
				{
					lBcc+=size;
					pos+=size;
					if(tmpbuffer[size-1]!='\n') {error=TRUE;goto endofdata;}
					if(!fgets(tmpbuffer,1000,stdin)) goto endofdata;
					fputs(tmpbuffer,stream);
					size=strlen(tmpbuffer);
				}
				while (tmpbuffer[0]==' ' || tmpbuffer[0]=='\t');
			}
			else if(!strncasecmp(tmpbuffer,"Subject:",8))
			{
				lSubject=-8;
				iSubject=pos+8;
				do
				{
					lSubject+=size;
					pos+=size;
					if(tmpbuffer[size-1]!='\n') {error=TRUE;goto endofdata;}
					if(!fgets(tmpbuffer,1000,stdin)) goto endofdata;
					fputs(tmpbuffer,stream);
					size=strlen(tmpbuffer);
				}
				while (tmpbuffer[0]==' ' || tmpbuffer[0]=='\t');
			}
			else
			{
				pos+=size;
				if(tmpbuffer[size-1]!='\n') {error=TRUE;goto endofdata;}
				if(!fgets(tmpbuffer,1000,stdin)) break;
				fputs(tmpbuffer,stream);
				size=strlen(tmpbuffer);
			}
		}
endofdata:
	}
	fclose(stream);

	if(error)
	{
		fault("Header line too long! More than 1000 chars!",FAULT_KILL);
		printf("Header line too long! Your mail client seems to be faulty!\n");
		return(100);
	}


/*-------------------------- Read Config file -----------------------------------------------------*/

   sprintf(tmpbuffer,"%s/.spam-config",home);
   if ((stream=fopen(tmpbuffer,"r")))
   {
      while(fgets(tmpbuffer,200,stream))
      {
         if(!strncasecmp(tmpbuffer,"DEBUG",5))
         {
            if(tmpbuffer[5]=='=')
            {
               if((tmpbuffer[6]=='y') || (tmpbuffer[6]=='Y')) debug=TRUE;
               if((tmpbuffer[6]=='n') || (tmpbuffer[6]=='N')) debug=FALSE;
            }
         }
         else if(!strncasecmp(tmpbuffer,"BOUNCE",6))
         {
            if(tmpbuffer[6]=='=')
            {
               if((tmpbuffer[7]=='y') || (tmpbuffer[7]=='Y')) bounce=TRUE;
               if((tmpbuffer[7]=='n') || (tmpbuffer[7]=='N')) bounce=FALSE;
            }
         }
         else if(!strncasecmp(tmpbuffer,"SHORTLOGS",9))
         {
            if(tmpbuffer[9]=='=')
            {
               if((tmpbuffer[10]=='y') || (tmpbuffer[10]=='Y')) shortlogs=TRUE;
               if((tmpbuffer[10]=='n') || (tmpbuffer[10]=='N')) shortlogs=FALSE;

               if(!strncmp(&tmpbuffer[10],"DISABLE",7))			/* Disable all logging. Only recommended */
					{																/* for emergency use */
						shortlogs=TRUE;
						nologs=TRUE;
					}
            }
         }
      }
      fclose(stream);
   }

/*--------------------------------------------------------------------------------------------------*/


	from=GetHeader(iFrom,lFrom);
	to=GetHeader(iTo,lTo);
	cc=GetHeader(iCc,lCc);
	bcc=GetHeader(iBcc,lBcc);

	subject=GetSubject();


	if((!from) && (!sender))
	{
		fault("No sender!",FAULT_KILL);
		FreeAll();
		if(debug)
			return(0);
		else
			return(99);
	}

	if(from) StrToLower(from);
	if(subject) StrToLower(subject); /* added ali 02-12-01 13:32*/

	if (to)
	{
		StrToLower(to);
		if(strstr(to,recipient)) dir=1;
	}
	if (cc)
	{
		StrToLower(cc);
		if(strstr(cc,recipient)) dir=1;
	}

	sprintf(tmpbuffer,"%s/.isdirect",home);
	if ((stream=fopen(tmpbuffer,"r")))
	{
		while(fgets(tmpbuffer,1000,stream) && (!dir))
		{
			if((tmpbuffer[0]!='#') && (tmpbuffer[0]!='\n') && (tmpbuffer[0]!='\r'))
			{
				if((tmpptr=strpbrk(tmpbuffer,"\n\r"))) *tmpptr=0x00;

				StrToLower(tmpbuffer);

				if (to) if(strstr(to,tmpbuffer)) dir=1;
				if (cc && !dir) if(strstr(cc,tmpbuffer)) dir=1;
			}
		}
		fclose(stream);
	}

	if (!dir) /* Not directly addressed! */
	{
		
		if((CheckAllowFile()) && (!CheckDenyFile())) /*added ali 02-12-01 11:59*/  
		/*if(CheckAllowFile()) removed ali 02-12-01 11:55*/
		{
			if(shortlogs)
			{
			   remove(tempfilename);
			}
			else fault("This was not directly addressed, but is allowed by user!",FAULT_NOTE);
		}
		else
		{
			fault("Not directly addressed!",FAULT_KILL);
			FreeAll();

			if(debug)
			{
				return(0);
			}
			else if(bounce)
			{
				PrintBounce();
				return(100);
			}
			else return(99);
		}
	}
	else		/* Directly addressed! */
	{
		if((CheckDenyFile()) && (!CheckAllowFile()))	/* Kolla bde deny och allow fr bttre matchning */
		{
			fault("This was directly addressed, but it is denied by user!",FAULT_KILL);
			FreeAll();
			if(debug)
			{
				return(0);
			}
			else if(bounce)
			{
				PrintBounce();
				return(100);
			}
			else return(99);
		}
		else if(shortlogs)
		{
		   remove(tempfilename);
		}
		else fault("",FAULT_NOERR);
	}

	FreeAll();
	return(0);
}

/*---------------------------- FreeAll() -----------------------------------------*/
void FreeAll(void)
{
	free(from);
	free(to);
	free(cc);
	free(bcc);
	free(subject);
	free(sender);
	free(recipient);
}



/*-------------------------- Print bounce message ----------------------------------------------*/
void PrintBounce(void)
{
	static char tmpbuffer[1000];		/* Potential security hazard.. If the user has a path longer than 1000 chars.. */
	FILE *stream;

	sprintf(tmpbuffer,"%s/.spam-bouncemessage",home);
   if ((stream=fopen(tmpbuffer,"r")))
   {
      while(fgets(tmpbuffer,200,stream))
      {
			printf("%s",tmpbuffer);
      }
      fclose(stream);
		printf("\n");
   }
	else printf("Your address is being denied at this time!\n");
}

/*---------------------------- StrToLower() --------------------------------------*/
void StrToLower(char *string)
{
	int a;
	for(a=0;string[a]!=0x00;a++)
	{
		string[a]=tolower(string[a]);
	}
}

/*---------------------------- CheckDenyFile() --------------------------------------*/
int CheckDenyFile(void)
{
   int deny=FALSE;
   char *tmpptr;
   FILE *stream;
   static char denyaddr[200];

   sprintf(denyaddr,"%s/.deny",home);
   if ((stream=fopen(denyaddr,"r")))
   {
      while(fgets(denyaddr,200,stream))
      {
         /* Skip blank lines and lines beginning with a '#' sign */
         if ((denyaddr[0]!='\r') && (denyaddr[0]!='\n') && (denyaddr[0]!='#'))
         {
				if((tmpptr=strpbrk(denyaddr,"\n\r"))) *tmpptr=0x00;

				StrToLower(denyaddr);

				if(from) if(strstr(from,denyaddr)) deny=TRUE;
				if(sender) if(strstr(sender,denyaddr)) deny=TRUE;
				if(to) if(strstr(to,denyaddr)) deny=TRUE;    /*added ali 02-12-01 12:16*/
                                if(cc) if(strstr(cc,denyaddr)) deny=TRUE;    /*added ali 02-12-01 12:16*/
				if(subject) if(CheckSubjectFile()) deny=TRUE;/*added ali 02-12-01 13:19*/
				
				if(deny)
				{
					fclose(stream);
					return(5);
				}
			}
      }
      fclose(stream);
   }
   return(0);
}

/*---------------------------- CheckAllowFile() --------------------------------------*/
int CheckAllowFile(void)
{
	int allow=FALSE;
	char *tmpptr;
	FILE *stream;
	static char allowaddr[200];

	sprintf(allowaddr,"%s/.allow",home);
	if ((stream=fopen(allowaddr,"r")))
	{
		while(fgets(allowaddr,200,stream))
		{
			/* Skip blank lines and lines beginning with a '#' sign */
			if ((allowaddr[0]!='\r') && (allowaddr[0]!='\n') && (allowaddr[0]!='#'))
			{
				if((tmpptr=strpbrk(allowaddr,"\n\r"))) *tmpptr=0x00;

				StrToLower(allowaddr);

				if(from) if(strstr(from,allowaddr)) allow=TRUE;
				if(to) if(strstr(to,allowaddr)) allow=TRUE;
				if(cc) if(strstr(cc,allowaddr)) allow=TRUE;
				if(sender) if(strstr(sender,allowaddr)) allow=TRUE;
				
				if(allow)
				{
					fclose(stream);
					return(5);
				}
			}
      }
      fclose(stream);
   }
   return(0);
}
/*------------------------------ CheckSubjectFile() -Added ali 02-12-01 13:17---------*/
int CheckSubjectFile(void)
{
	int subjectBlocked=FALSE;
	char *tmpptr;
	FILE *stream;
	static char subjectaddr[200];

	sprintf(subjectaddr,"%s/.subject",home);
	if ((stream=fopen(subjectaddr,"r")))
	{
		while(fgets(subjectaddr,200,stream))
		{
			/* Skip blank lines and lines beginning with a '#' sign */
			if ((subjectaddr[0]!='\r') && (subjectaddr[0]!='\n') && (subjectaddr[0]!='#'))
			{
				if((tmpptr=strpbrk(subjectaddr,"\n\r"))) *tmpptr=0x00;

				StrToLower(subjectaddr);
				if(subject) if(strstr(subject,subjectaddr)) subjectBlocked=TRUE;

				if(subjectBlocked)
				{
					fclose(stream);
					return(5);
				}
			}
      }
      fclose(stream);
   }
   return(0);
}
/*------------------------------ GetHeader() ------------------------------------------*/
char * GetHeader(int index,int length)
{
	unsigned char *memptr,*memptr2,*tmpptr;
	int tmp,tmp2,len;
	FILE *stream;

	if(index==0||length==0) return(0);
	if(!(memptr=(char *)malloc(length))) return(0);

	if (!(stream=fopen(tempfilename,"r"))) return(0);

	fseek(stream,index,SEEK_SET);
	fread(memptr,length,1,stream);
 	fclose(stream);

	for(tmp=0,len=0;tmp<length;tmp++)
	{
		if(memptr[tmp]=='(')
		{
			if((tmpptr=strchr(&memptr[tmp],')'))) tmp += (int) (tmpptr - &memptr[tmp]);
		}
		else if(memptr[tmp]=='[')
		{
			if((tmpptr=strchr(&memptr[tmp],']'))) tmp += (int) (tmpptr - &memptr[tmp]);
		}
		else if((memptr[tmp]>0x20) && (memptr[tmp]<0x80))  /* Skip spaces and other illegal chars */
		{
			len++;
		}
	}

	if(!(memptr2=(char *)malloc(len+1)))
	{
		free(memptr);
		return(0);
	}

	for(tmp=0,tmp2=0;tmp<length;tmp++)
	{
		if(memptr[tmp]=='(')
		{
			if((tmpptr=strchr(&memptr[tmp],')'))) tmp += (int) (tmpptr - &memptr[tmp]);
		}
		else if(memptr[tmp]=='[')
		{
			if((tmpptr=strchr(&memptr[tmp],']'))) tmp += (int) (tmpptr - &memptr[tmp]);
		}
		else if((memptr[tmp]>0x20) && (memptr[tmp]<0x80))  /* Skip spaces and other illegal chars */
		{
			memptr2[tmp2++]=memptr[tmp];
		}
	}
	memptr2[tmp2]=0x00;

	free(memptr);
	return(memptr2);
}


/*------------------------------ GetSubject() ------------------------------------------*/
char * GetSubject(void)
{
	unsigned char *memptr,*tmpptr;
	FILE *stream;

	if(iSubject==0||lSubject==0) return(0);
	if(!(memptr=(char *)malloc(lSubject+1))) return(0);

	if (!(stream=fopen(tempfilename,"r"))) return(0);

	fseek(stream,iSubject,SEEK_SET);
	fread(memptr,lSubject,1,stream);
 	fclose(stream);

	do { if((tmpptr=strpbrk(memptr,"\n\r"))) *tmpptr=' '; } while (tmpptr);

	memptr[lSubject]=0x00;

	return(memptr);
}
/*-------------------------------- fault() ------------------------------------------*/
void fault(char *err,int errtype)
{
	FILE *stream,*tmpstream;
	char buffer[500];
	int size;

	strftime(tstamp,80, "Mail received on: %y-%m-%d, at %H:%M:%S\n",localtime(&curtime));

	if (ext[0]==0x00)
      sprintf(buffer,"%s/.maillog",home);
   else
      sprintf(buffer,"%s/.maillog-%s",home,ext);

	if(!nologs)
	{
		if ((stream=fopen(buffer,"a")))
		{
			fprintf(stream,"%s",tstamp);
			fprintf(stream,"Sender: %s\n",sender);
			if(from) fprintf(stream,"From: %s\n",from);
			if(to) fprintf(stream,"To: %s\n",to);
			if(cc) fprintf(stream,"Cc: %s\n",cc);
			if(bcc) fprintf(stream,"Bcc: %s\n",bcc);
			if (subject) fprintf(stream,"Subject: %s\n",subject);
			if (debug) fprintf(stream,"lFrom: %d, lTo: %d, lCc: %d, lBcc: %d, lSubject: %d\n",lFrom,lTo,lCc,lBcc,lSubject);

			switch(errtype)
			{
				case FAULT_NOERR:
					fprintf(stream,"Accepted!\n\n");
					break;
				case FAULT_NOTE:
					fprintf(stream,"Accepted! (%s)\n\n",err);
					break;
				case FAULT_ERR:
					fprintf(stream,"Error! (%s)\n\n",err);
					break;
				case FAULT_KILL:
					fprintf(stream,"Rejected! (%s)\n\n",err);
					break;
			}

			fclose(stream);
		}
		if(errtype==FAULT_KILL)								/* Save killed mail in file */
		{
			if ((tmpstream=fopen(tempfilename,"r")))
			{
				sprintf(buffer,"%s/.killedmails",home);
				if ((stream=fopen(buffer,"a")))
				{
					strftime(tstamp,80, "%a %b %d %H:%M:%S %Y",localtime(&curtime));
					fprintf(stream,"From %s %s\n",recipient,tstamp);
					
					while((size=fread(buffer,1,500,tmpstream)))	{fwrite(buffer,size,1,stream);}
					while((size=fread(buffer,1,500,stdin)))		{fwrite(buffer,size,1,stream);}

					fprintf(stream,"\n\n");
					fclose(stream);
				}
				fclose(tmpstream);
			}
		}
	}
   remove(tempfilename);
}

